
#ifndef GCODER_H
#define GCODER_H 
 
/* ----------------------------------------------------------------------
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  Purpose: 
  Version: $Rev:$

  2011-04-20 /JD: created
  2013-03-07 /JD: general settings supported

-*/

#include <set>
#include <map>
#include <vector>
#include <string>
#include <cstring>
#ifdef BMUTEX
#include <boost/thread/mutex.hpp>
#endif

#include "gdata/gdata.h"
#include "gutils/gsys.h"
#include "gutils/gmutex.h"
#include "gset/gsetbase.h"
#include "gset/gsetgnss.h"
#include "gset/gsetgen.h"

#define BUFFER_INCREASE_FAC  1.5
#define DEFAULT_BUFFER_SIZE  4096
#define MAXIMUM_BUFFER_SIZE  240000 // 9000000

namespace gnut {

// enum t_irc { failure = -1, success, fatal } ; // return code

static const char lf = 0x0a; // = "\n" .. line feed
static const char cr = 0x0d; // = "\r" .. carriage return

#if defined __linux__ || defined __APPLE__
static const string crlf = string("") + lf;   // "\n"
#endif

#if defined _WIN32 || defined _WIN64
//static const string crlf = string("") + cr + lf; // "\r\n"
static const string crlf = string("") + lf; // "\r\n"
#endif 

using namespace std;

class t_gcoder {

 public:
   t_gcoder( string version = "", int sz = DEFAULT_BUFFER_SIZE );
   virtual ~t_gcoder();

   virtual void clear();

   virtual void version(string s){ _version = s; }
   virtual string version(){ return _version; }

   virtual void out_length(int len){ _out_length = len; }
   virtual int  out_length(){ return _out_length; }
   
   virtual void  out_sample(int smp){ _out_sample = smp; }
   virtual float out_sample(){ return _out_sample; }
   
   virtual void    out_epoch(t_gtime epo){ _out_epoch = epo; }
   virtual t_gtime out_epoch(){ return _out_epoch; }

   virtual int decode_head( char* buff, int sz,           vector<string>& errmsg ){ return 0; } // = 0;
   virtual int decode_data( char* buff, int sz, int& cnt, vector<string>& errmsg ){ return 0; } // = 0;

   virtual int encode_head( char* buff, int sz,           vector<string>& errmsg ){ return 0; } // = 0;
   virtual int encode_data( char* buff, int sz, int& cnt, vector<string>& errmsg ){ return 0; } // = 0;

   virtual int gset(t_gsetbase* s);              // get settings
   void glog(t_glog* l){ _log = l; }             // set/get glog pointer
   t_glog* glog(){ return _log; }

   int add_data( string data_id, t_gdata* data );
   int rem_data( string data_id);
   t_gdata* data( string data_id ){ return _data[data_id]; }

 protected:
   int   endpos(){ return _endpos; }
   int     size(){ return _endpos; }             // number of char elements in buffer (excl \0)
   char* buffer(){ return _buffer; }

   virtual void _init();
   int _getline(string& str, int from_pos = 0);
   int _add2buffer( char* buff, int sz );
   int _consume( int bytes_to_eat );
    
   bool                    _initialized;         // if initialized
   int                     _out_length;          // encoder data batch length [min]
   float                   _out_sample;          // encoder data batch sample [sec]
   t_gtime                 _out_epoch;           // encoder data refrence epoch
   string                  _version;             // format version
   string                  _initver;             // format initial version
   map<string,t_gdata*>    _data;                // data pointer
   t_glog*                 _log;                 // log pointer
   t_gsetbase*             _set;                 // set pointer
   char*                   _buffer;              // class buffer
   int                     _buffsz;              // size of buffer
   int                     _endpos;              // after last position
   
   // settings
   t_gtime     _beg;                             // default beg time
   t_gtime     _end;                             // default end time
   double      _int;                             // default interval
   set<string> _sys;                             // default systems
   set<string> _rec;                             // default sites/receivers
   
   map<GSYS,set<string>> _sat;                   // default satellites
   map<GSYS,set<string>> _obs;                   // default observations (all mixed for single GNSS !!)

  // ENCODING
   int _fill_buffer( char* buff, int sz );
   stringstream     _ss;
   long             _ss_position;

 
#ifdef BMUTEX
   boost::mutex            _mutex;               // buffer mutex
#else
  t_gmutex                 _mutex;               // buffer mutex
#endif
 private:

};

} // namespace

#endif
