
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <iomanip>
#include <sstream>
#include <algorithm>

#include "gset/gsetinp.h"

using namespace std;
using namespace pugi;

namespace gnut {

// Convertor for INP formats
// ----------
IFMT t_gsetinp::str2ifmt(const string& s)
{ 
  string tmp = s;
  transform(tmp.begin(), tmp.end(), tmp.begin(), ::toupper);
  if( tmp == "RINEXC"  ) return RINEXC_INP;
  if( tmp == "RINEXO"  ) return RINEXO_INP;
  if( tmp == "RINEXN"  ) return RINEXN_INP;
  if( tmp == "BNCOBS"  ) return BNCOBS_INP;
  if( tmp == "BNCRTCM" ) return BNCRTCM_INP; 
  if( tmp == "TROSINEX") return TROSINEX_INP;
  if( tmp == "RTCM"    ) return RTCM_INP;
  if( tmp == "TSMET"   ) return TSMET_INP;
  if( tmp == "SP3"     ) return SP3_INP;
  if( tmp == "ATX"     ) return ATX_INP;
  if( tmp == "BLQ"     ) return BLQ_INP;   
  if( tmp == "GRIB"    ) return GRIB_INP;   
  if( tmp == "NCD3"    ) return NCD3_INP;   
  if( tmp == "GRID"    ) return GRID_INP;   
   
  return IFMT(-1);
}

// Convertor for INP formats
// ----------
string t_gsetinp::ifmt2str(const IFMT& f)
{ 
  switch ( f ){        
   case RINEXO_INP:  return "RINEXO";
   case RINEXC_INP:  return "RINEXC";
   case RINEXN_INP:  return "RINEXN";
   case BNCOBS_INP:  return "BNCOBS";
   case BNCRTCM_INP: return "BNCRTCM";
   case TROSINEX_INP:return "TROSINEX";
   case RTCM_INP:    return "RTCM";
   case TSMET_INP:   return "TSMET";
   case SP3_INP:     return "SP3";
   case ATX_INP:     return "ATX";
   case BLQ_INP:     return "BLQ";     
   case GRIB_INP:    return "GRIB";
   case NCD3_INP:    return "NCD3";
   case GRID_INP:    return "GRID";
   default:          return "UNDEF";
  }
  return "UNDEF";
}

// Constructor
// ----------
t_gsetinp::t_gsetinp()
  : t_gsetbase()
{}


// Destructor
// ----------
t_gsetinp::~t_gsetinp()
{}


// Get formats input size
// ----------
int t_gsetinp::input_size(const string& fmt)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  int tmp = _inputs(fmt).size();

  _gmutex.unlock(); return tmp;
}


// Get formats inputs (all in multimap)
// ----------
multimap<IFMT,string> t_gsetinp::inputs_all()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();
   
  multimap<IFMT,string> map;

  set<string> ifmt = _iformats();
  set<string>::const_iterator itFMT = ifmt.begin();
  while( itFMT != ifmt.end() )
  {
    string fmt = *itFMT;
    IFMT  ifmt = str2ifmt( fmt );
    set<string> inputs = _inputs( fmt );
    set<string>::const_iterator itINP = inputs.begin();
    while( itINP != inputs.end() )
    {
      map.insert(map.end(), pair<IFMT,string>( ifmt, *itINP ));
      itINP++;
    }
    itFMT++;
  }
  _gmutex.unlock(); return map;
}


// Get formats inputs
// ----------
set<string> t_gsetinp::inputs(const string& fmt)
{
   return _inputs(fmt);
}


// Get input formats
// ----------
set<string> t_gsetinp::iformats()
{
   return _iformats();  
}


// Get formats inputs
// ----------
set<string> t_gsetinp::_inputs(const string& fmt)
{
  set<string> tmp;
  string str;

  for( xml_node node = _doc.child(XMLKEY_ROOT).child(XMLKEY_INP).first_child(); node; node = node.next_sibling() ){
    if( node.name() == fmt ){       
      istringstream is( node.child_value() );
      while( is >> str && !is.fail() ){
        if( str.find("://") == string::npos ) str = "file://" + str;
	tmp.insert( str );
      }
    }
  }
  return tmp;
}


// Get input formats
// ----------
set<string> t_gsetinp::_iformats()
{
  set<string> tmp;
  for( xml_node node = _doc.child(XMLKEY_ROOT).child(XMLKEY_INP).first_child(); node; node = node.next_sibling() ){
    tmp.insert( node.name() );
  }
  return tmp;
}


// settings check
// ----------
void t_gsetinp::check()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif     
  _gmutex.lock();
   
  // check existence of nodes/attributes
  xml_node parent = _doc.child(XMLKEY_ROOT);
  xml_node node = _default_node(parent, XMLKEY_INP);

  // check supported input formats (see IFMT enum !)
  set<string> ifmt = _iformats();
  set<string>::const_iterator itFMT = ifmt.begin();
  while( itFMT != ifmt.end() ){
    string fmt = *itFMT;
    IFMT  ifmt = str2ifmt( fmt );
    if( ifmt < 0 ){
      _doc.child(XMLKEY_ROOT).child(XMLKEY_INP).remove_child( node.child(fmt.c_str()) );
      if( _log ) _log->comment(0, "Warning: " + fmt + " inp format not implemented [gsetinp::check()]!");
      itFMT++;
      continue;
    }
  
    // check application-specific output format
    if( _IFMT_supported.find( ifmt ) == _IFMT_supported.end() ){
      _doc.child(XMLKEY_ROOT).child(XMLKEY_INP).remove_child( node.child(fmt.c_str()) );
      if( _log ) _log->comment(0, "Warning: " + fmt + " inp format not supported by this application!");
      else                cerr << "Warning: " + fmt + " inp format not supported by this application!\n";       
    }
    itFMT++;            
  }
#ifdef BMUTEX
  lock.unlock();
#endif
   _gmutex.unlock(); return;
}


// settings help
// ----------
void t_gsetinp::help()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif   
  _gmutex.lock();
   
  
  cerr << " <inputs>\n"
       << "   <rinexo> file://dir/name </rinexo> \t\t <!-- obs RINEX decoder -->\n"
       << "   <rinexn> file://dir/name </rinexn> \t\t <!-- nav RINEX decoder -->\n"
       << " </inputs>\n";

  cerr << "\t<!-- inputs description:\n"
       << "\t <decoder> path1 path2 path3  </decoder>\n"
       << "\t ... \n"
       << "\t where path(i) contains [file,tcp,ntrip]:// depending on the application\n"
       << "\t -->\n\n";

#ifdef BMUTEX
  lock.unlock();
#endif
   _gmutex.unlock(); return;  
}

} // namespace
